/* ----------------------------------------------------------------------------
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 * ----------------------------------------------------------------------------
 */

#include <stdio.h>

/* rdmsql.h is required in all files using RDMSQL */
#include "rdmsql.h"

/* EXAMPLE - SQL13 for C using RDMSQL */
/* This example shows multi-user access to a database using separate processes
 * and a remote tfserver.  The connections will use TCP/IP to communicate
 * with the tfserver to do their work.  After starting the tfserver, two
 * writer processes are created.  The first begins inserting values into the
 * table.  The second waits until the table contains more than 100 rows and
 * then begins deleting the oldest.  A reader process is also created that
 * prints out the current oldest and newest entries every second.  Each
 * process uses regular locking.
 *
 * The following RDM functions are illustrated:
 *      rsqlAllocConn/rsqlFreeConn
 *      rsqlAllocStmt/rsqlFreeStmt
 *      rsqlOpenDB/rsqlCloseDB
 *      rsqlExecDirect
 *      rsqlFetch
 *      rsqlPrepare
 *      rsqlBindNamedParam
 *      rsqlExecute
 *      rsqlSetAutoCommit
 *      d_tfsinitEx/d_tfstermEx
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 * 
 * For simplicity, this example does not check all return codes, but good
 * programming practices would dictate that they are checked after each
 * RDMSQL call.
 */
static void drop_database(
    HCONN       hDbc,
    const char *name)
{
    HSTMT hStmt;
    char  commandText[80];

    (void) rsqlAllocStmt(hDbc, &hStmt);

    /* Attempt to drop the database.  If the database does not already exist
     * this will fail and we simply ignore the failure for this sample. */
    sprintf(commandText, "DROP DATABASE %s", name);
    (void) rsqlExecDirect(hStmt, commandText);
    (void) rsqlExecDirect(hStmt, "COMMIT");

    (void) rsqlFreeStmt(hStmt);
}

static void create_this_database(
    HCONN       hDbc,
    const char *name,
    const char *cmds[],
    uint32_t    cmd_length)
{
    RSQL_ERRCODE rsqlCode;
    HSTMT        hStmt;
    char         commandText[80];
    uint16_t     ii;

    (void) rsqlAllocStmt(hDbc, &hStmt);

    sprintf(commandText, "CREATE DATABASE %s", name);
    rsqlCode = rsqlExecDirect(hStmt, commandText);

    for (ii = 0; ii < cmd_length; ii++)
    {
        rsqlCode = rsqlExecDirect(hStmt, cmds[ii]);
        if (rsqlCode != errSUCCESS)
            break;
    }

    if (rsqlCode != errSUCCESS)
        fprintf(stderr, "Error creating the %s database\n", name);

    (void) rsqlExecDirect(hStmt, "COMMIT");

    /* Close the newly created database */
    (void) rsqlCloseDB(hDbc, name);

    (void) rsqlFreeStmt(hStmt);
}

static void create_databases(
    HCONN hDbc)
{
    const char *cmds[] = {
        "CREATE TABLE Table1 (id INTEGER PRIMARY KEY)"
    };

    /* Drop any existing copies of the databases. */
    drop_database(hDbc, "SQL13_db");

    /* Attempt to create the databases */
    create_this_database(hDbc, "SQL13_db", cmds, 1);
}

static void inserter(void)
{
    HCONN   hDbc;
    HSTMT   hStmt;
    int32_t param_id;
    int32_t ii = 0;

    printf("Inserter process\n");

    /* Create a connection for this thread */
    (void) rsqlAllocConn(&hDbc);
    (void) rsqlOpenDB(hDbc, "SQL13_db", "s");

    /* Set AUTO COMMIT on */
    (void) rsqlSetAutoCommit(hDbc, 1);

    /* Create a statement to do the insert */
    (void) rsqlAllocStmt(hDbc, &hStmt);

    rsqlPrepare(hStmt, "INSERT INTO Table1 VALUES :id");
    rsqlBindNamedParam(hStmt, "id", tINTEGER, &param_id, NULL, NULL);

    /* Loop until a shutdown is requested */
    while (1)
    {
        /* Insert th next value */
        param_id = ii++;
        (void) rsqlExecute(hStmt);
    }
}

static void deleter(void)
{
    RSQL_ERRCODE      rsqlCode;
    HCONN             hDbc;
    HSTMT             hChkStmt;
    HSTMT             hMinStmt;
    HSTMT             hDelStmt;
    int32_t           param_id;
    const RSQL_VALUE *chkResult;
    const RSQL_VALUE *minResult;
    uint16_t          numCols;

    printf("Deleter process\n");

    /* Create a connection for this thread */
    (void) rsqlAllocConn(&hDbc);
    (void) rsqlOpenDB(hDbc, "SQL13_db", "s");

    /* Set AUTO COMMIT on */
    (void) rsqlSetAutoCommit(hDbc, 1);

    /* Create a command to check the number of rows, one to fine the 'smallest'
     * row, and one to delete */
    (void) rsqlAllocStmt(hDbc, &hChkStmt);
    (void) rsqlAllocStmt(hDbc, &hMinStmt);
    (void) rsqlAllocStmt(hDbc, &hDelStmt);

    rsqlCode = rsqlPrepare(hChkStmt, "SELECT COUNT(id) FROM Table1");
    rsqlCode = rsqlPrepare(hMinStmt, "SELECT MIN(id) FROM Table1");
    rsqlCode = rsqlPrepare(hDelStmt, "DELETE FROM Table1 WHERE id = :minid");
    rsqlCode = rsqlBindNamedParam(hDelStmt, "minid", tINTEGER, &param_id, NULL, NULL);

    /* Loop until a shutdown is requested */
    while (1)
    {
        /* Query the number of rows in the table */
        (void) rsqlExecute(hChkStmt);
        rsqlCode = rsqlFetch(hChkStmt, &chkResult, &numCols);
        if (rsqlCode == errSUCCESS)
        {
            int32_t count = chkResult[0].vt.lv;

            /* There should be only one row, so close the statement */
            (void) rsqlCloseStmt(hChkStmt);

            if (count > 100)
            {
                /* If there at least 100 rows, get the smallest and delete it */
                rsqlCode = rsqlExecute(hMinStmt);
                rsqlCode = rsqlFetch(hMinStmt, &minResult, &numCols);
                if (rsqlCode == errSUCCESS)
                {
                    param_id = minResult[0].vt.lv;
                    rsqlCode = rsqlExecute(hDelStmt);

                    /* There should be only one row, so close the statement */
                    (void) rsqlCloseStmt(hMinStmt);
                }
            }
        }
    }
}

static void reader(void)
{
    RSQL_ERRCODE      rsqlCode;
    HCONN             hDbc;
    HSTMT             hStmt;
    char              commandText[80];
    const RSQL_VALUE *result;
    uint16_t          numCols;

    printf("Reader process\n");

    /* Create a connection for this thread */
    (void) rsqlAllocConn(&hDbc);
    (void) rsqlOpenDB(hDbc, "SQL13_db", "s");

    (void) rsqlAllocStmt(hDbc, &hStmt);

    sprintf(commandText, "SELECT MIN(id), MAX(id) FROM Table1");

    /* Loop until a shutdown is requested */
    while (1)
    {
        /* Sleep one second */
        sleep(1);

        rsqlCode = rsqlExecDirect(hStmt, commandText);
        if (rsqlCode == errSUCCESS)
        {
            rsqlCode = rsqlFetch(hStmt, &result, &numCols);
            if (rsqlCode == errSUCCESS)
            {
                printf("Current minimum = %d and current maximum = %d\n",
                        result[0].vt.sv, result[1].vt.sv);

                (void) rsqlCloseStmt(hStmt);
            }
        }
    }
}

/* =======================================================================
*/
static pid_t start_tfserver(void)
{
    pid_t pid = fork();
    if (pid < 0)
    {
        fprintf(stderr, "Error forking (%s)\n", strerror(errno));
        exit(EXIT_FAILURE);
    }

    if (pid == 0)
    {
        int32_t     stat;
        const char *tfserver_main_argv[] = {
            "tfserver"
        };

        (void) psp_init();

        stat = tfserver_main(1, tfserver_main_argv);
        if (stat != S_OKAY)
        {
            fprintf(stderr, "Error starting tfserver (%d)\n", stat);
            psp_term();
            exit(EXIT_FAILURE);
        }
    }

    return pid;
}

/* =======================================================================
*/
static void stop_process(
    pid_t pid)
{
    int ret;
    int stat;

    /* Kill the tfserver process */
    ret = kill(pid, SIGINT);
    if (ret == -1)
    {
        fprintf(stderr, "Error killing tfserver (%s)\n", strerror(errno));
        exit(EXIT_FAILURE);
    }

    /* Wait for the tfserver process to terminate */
    do
    {
        waitpid(pid, &stat, 0);
    } while (!WIFEXITED(stat) && !WIFSIGNALED(stat));
}

/* =======================================================================
*/
int32_t main(
    int32_t            argc,
    const char *const *argv)
{
    HCONN        hDbc;
    TFS_HANDLE   hTFS;
    pid_t        pid;
    pid_t        pid_i;
    pid_t        pid_d;
    pid_t        pid_r;

    printf("Starting tfserver (and waiting 5 seconds for it to initialize)\n");

    /* Start the tfserver process to process 'remote' connections */
    pid = start_tfserver();
    if (pid > 0)
    {
        /* Sleep to give the tfserver time to initialize and start listening */
        sleep(5);

        /* Specify that we use a remote tfserver */
        (void) d_tfsinitEx(NULL, NULL, &hTFS, TFS_TYPE_RPC);

        /* Create the first connection */
        (void) rsqlAllocConn(&hDbc);

        create_databases(hDbc);

        /* Create three processes one to insert, one to delete, and one to show the
         * current contents */
        printf("Starting threads\n");
        printf("   Press <Enter> when you are ready to terminate them\n");

        pid_i = fork();
        if (pid_i == 0)
            inserter();
        else if (pid_i > 0)
        {
            pid_d = fork();
            if (pid_d == 0)
                deleter();
            else if (pid_d > 0)
            {
                pid_r = fork();
                if (pid_r == 0)
                    reader();
                else if (pid_r > 0)
                {
                    /* Wait for user to hit <Enter> */
                    getchar();

                    rsqlFreeConn(hDbc);
                    d_tfsterm(hTFS);
                    
                    stop_process(pid_r);
                }
                stop_process(pid_d);
            }
            stop_process(pid_i);
        }

        /* Kill the tfserver process */
        stop_process(pid);
    }

    return EXIT_SUCCESS;
}

